(*
 * KGL Abstract Syntax Tree types 
 *
 *)


(* Binary operators *)
type op = Plus  | Minus  | Multi | Divide | Mod 
        | Equal | Nequal | Less  | Leq    | Greater | Geq 
        | In    | And    | Or

(* Assignment operators *)
type asnop = Add | Sub | Asn


(* Expressions *)
type expr = 
    IntLit of int
  | FloatLit of float
  | BoolLit of bool
  | CharLit of char
  | StrLit of string
  | Id of string
  | Assign of expr * asnop * expr  (* Assignment: expr1 =/+=/-= expr2 *) 
  | Not of expr 
  | Binop of expr * op * expr 
  | Call of string * expr list     (* Function call: f(param1, param2,...) *)
  | Value of expr * expr           (* Reference: expr1[expr2] *)
  | ListLit of expr list           (* List literal: [| e1, e2, e3, ... |] *)
  | DictLit of kv list             (* Dict literal: (| k1:v1, k2:v2, ... |) *)
  | GraphLit of edge list          (* Graph literal: {|src--(label)-->dest; ...|}*)
  | Null
  | Noexpr 

and kv = {
        key: expr;
        value: expr; 
}

and edge = { 
        src: expr;     
        dest: expr;
        label: expr;
}


(* Variable Types and Declaration *)
type var_type =
          Int   | Float | Bool | Char | Str 
        | Graph | Node 
        | List of var_type
        | Dict of var_type * var_type
        | Void

type var_decl = {
        v_type: var_type;
        v_name: string;
        v_init: expr;
}

(* Statements *)
type stmt =
        | Block of stmt list
        | Variable of var_decl 
        | Expr of expr 
        | Return of expr
        | If of expr * stmt * stmt
        | For of expr * expr * expr * stmt
        | Foreach of expr * stmt
        | While of expr * stmt
        | Continue
        | Break


(* Function Declaration *)
type func_decl = {
        fname: string;
        formals: var_decl list;
        rtype: var_type;  
        body: stmt list;
}

type program = var_decl list * func_decl list

exception Variable_of_void



(*Prints Out the AST*)
let string_of_binop = function 
    Plus   -> "+"
  | Minus  -> "-"
  | Multi  -> "*"
  | Divide -> "/"
  | Mod    -> "%"
  | Equal  -> "=="
  | Nequal -> "!="
  | Less   -> "<"
  | Leq    -> "<="
  | Greater -> ">"
  | Geq    -> ">="
  | In     -> "in"
  | And    -> "&&"
  | Or     -> "||" 

let string_of_asnop = function
    Add    -> "+="
  | Sub    -> "-="
  | Asn    -> "="

let rec string_of_expr = function
    IntLit(lit)   -> "IntLit( " ^ string_of_int lit ^ " )"  
  | FloatLit(lit) -> "FloatLit( " ^ string_of_float lit ^ " )"
  | BoolLit(lit)  -> "BoolLit( " ^ string_of_bool lit ^ " )"
  | CharLit(lit)  -> "CharLit( " ^ Char.escaped lit ^ " )"
  | StrLit(lit)   -> "StrLit( " ^ lit ^ " )"
  | Id(id)        -> "Id( " ^ id ^ " )"
  | Assign(e1, asnop, e2) -> "Assign( " ^ string_of_expr e1 ^ " " ^ 
                             string_of_asnop asnop ^ " " ^ string_of_expr e2 ^ " )"
  | Not(expr)     -> "Not( " ^ string_of_expr expr ^ " )"
  | Binop(expr1,op,expr2) -> "Binop( " ^ string_of_expr expr1 ^ " " ^ string_of_binop op ^ " " ^ string_of_expr expr2 ^ " )"
  | Call(fname,param)  -> "Call( " ^ fname ^ ", " ^ String.concat "; " (List.map string_of_expr param) ^ " )" 
  | Value(expr1,expr2) -> "Value( " ^ string_of_expr expr1 ^ ", " ^ string_of_expr expr2 ^ " )"
  | ListLit(elements) -> "ListLit( [" ^ String.concat "; " (List.map string_of_expr elements) ^ "] )" 
  | DictLit(kvpairs) -> "DictLit( [" ^ String.concat ";" (List.map string_of_dict  kvpairs) ^ "] )"
  | GraphLit(edges) -> "GraphLit( [" ^ String.concat ";" (List.map string_of_edge edges) ^ "] )"
  | Null -> "Null"
  | Noexpr -> "Nonexpr"

and string_of_edge edge = string_of_expr edge.src ^ "-- (" ^ string_of_expr edge.label ^ ") -->" ^ string_of_expr edge.dest

and string_of_dict kvpair = string_of_expr kvpair.key ^ ":" ^ string_of_expr kvpair.value

let rec string_of_var_type = function
  Int    -> "int"
| Float  -> "float"
| Bool   -> "boolean"
| Char   -> "char"
| Str    -> "string"
| Graph  -> "graph"
| Node   -> "node"
| List(var_type)   -> "list(" ^ string_of_var_type var_type ^ ")"
| Void   -> "void"
| Dict(t1, t2)   -> "dict(" ^ string_of_var_type t1 ^ "," ^ string_of_var_type t2 ^ ")"
 
let string_of_var_decl vardecl = "Var( name: " ^ vardecl.v_name ^ "; " ^ string_of_var_type vardecl.v_type ^ "; " ^ string_of_expr vardecl.v_init ^ " )"

let rec string_of_stmt = function
    Block(stmt_list)   -> (String.concat "\n" (List.map string_of_stmt stmt_list)) ^ "\n"
  | Variable(var)      ->  "Variable( " ^ (string_of_var_decl var) ^ " )"
  | Expr(expr)         ->  "Expr( " ^ (string_of_expr expr) ^ " )"
  | Return(expr)       ->  "Return( " ^ (string_of_expr expr) ^ ")"  
  | If(expr,stmt1,stmt2) -> "if (" ^ (string_of_expr expr) ^ ")\n" ^ (string_of_stmt stmt1) ^ (string_of_stmt stmt2) ^ ")"
  | For(init, test, after, stmt) -> "for (" ^ string_of_expr init ^ ", " ^ string_of_expr test ^ ", " ^ string_of_expr after ^ ") {\n" ^ string_of_stmt stmt ^ "\n}"
  | Foreach(expr,stmt)   ->  "Foreach (" ^ (string_of_expr expr) ^ ") {\n" ^ (string_of_stmt stmt) ^ "\n}"
  | While(test, stmt) -> "while( " ^ (string_of_expr test) ^ " ) {\n" ^ (string_of_stmt stmt) ^ "\n}"
  | Continue          -> "continue"
  | Break             -> "break"

let string_of_func_decl funcdecl = "Function( type: (" ^ string_of_var_type funcdecl.rtype ^ ") name: \"" ^ funcdecl.fname ^ "\" formals: " 
  ^ (String.concat ", " (List.map string_of_var_decl funcdecl.formals))
  ^ ") {\n" ^ String.concat "\n\n" (List.map string_of_stmt funcdecl.body) ^ "\n}"

let string_of_program (var_list,func_list)= "Program_START\n" ^ (String.concat "\n" (List.map string_of_var_decl var_list)) ^ "\n\n" ^
  (String.concat "\n\n" (List.map string_of_func_decl (List.rev func_list) )) ^ "\nProgram_END\n"
  
