(*
 * KGL C++ AST and Code Generation 
 *)

open Ast
open Sast
exception Error of string


(* Types *)
type c_var_type =
        Int | Double | Bool | Char | Str 
      | GraphPtr (* Reference to graph *) 
      | NodePtr  (* Pointer to node *)
      | Vector of c_var_type
      | Map of c_var_type * c_var_type
      | Auto
      | Void  


(* SAST Type -> CAST Type *)
let rec to_c_var_type (t : Ast.var_type) : c_var_type = 
  match t with
    Ast.Int -> Int
  | Ast.Float -> Double
  | Ast.Bool -> Bool
  | Ast.Char -> Char
  | Ast.Str -> Str
  | Ast.List(v) -> Vector(to_c_var_type v)
  | Ast.Dict(v1, v2) -> Map(to_c_var_type v1, to_c_var_type v2)
  | Ast.Node -> NodePtr
  | Ast.Graph -> GraphPtr
  | Ast.Void -> Void 

let rec string_c_var_type (vt : c_var_type) = match vt with
    Int -> "int" 
  | Double -> "double"
  | Char -> "char"
  | Str -> "string"
  | Bool -> "bool" 
  | GraphPtr -> "Graph &"
  | NodePtr -> "Node *"
  | Vector(v) -> "vector<" ^ string_c_var_type v ^ ">"
  | Map(kt, vt) -> 
      "unordered_map<" ^ string_c_var_type kt ^ "," ^ string_c_var_type vt ^ ">"
  | Auto -> "auto"
  | Void -> "void" 


(* CAST variable declaration *)
type c_var_decl = {
  c_vtype: c_var_type;
  c_vname: string;
  c_vinit: c_expr;
}

(* CAST expr *)
and c_expr = 
    C_IntLit of int
  | C_DoubleLit of float
  | C_BoolLit of bool
  | C_CharLit of char
  | C_StrLit of string
  | C_Id of string
  | C_Assign of c_expr * asnop * c_expr
  | C_Not of c_expr
  | C_Binop of c_expr * op * c_expr
  | C_Call of string * (c_expr list)
  | C_Value of c_expr * c_expr
  | C_Null
  | C_Noexpr
  | C_Exprstmt of c_stmt list
  | C_ListLit of c_expr list
  | C_DictLit of (c_expr * c_expr) list

(* CAST stmt *)
and c_stmt = 
    C_Block of c_stmt list
  | C_Variable of c_var_decl 
  | C_Expr of c_expr
  | C_Return of c_expr
  | C_If of c_expr * c_stmt * c_stmt
  | C_For of c_expr * c_expr * c_expr * c_stmt
  | C_Foreach of c_expr * c_expr * c_stmt 
  | C_While of c_expr * c_stmt
  | C_Continue
  | C_Break

(* CAST function declaration *)
type c_func_decl = {
  c_fname: string;
  c_formals: (c_var_type * string) list;
  c_rtype: c_var_type;
  c_body: c_stmt list;
}

(* CAST program *)
type c_program = {
  c_global_vars: c_var_decl list;
  c_func_decls: c_func_decl list;
  headers: string list;
}

let rec to_c_var_decl (var: Sast.s_var_decl) : c_var_decl =
  let typ = to_c_var_type var.s_vtype and init = to_c_expr var.s_vinit in
  {
    c_vtype = typ;
    c_vname = var.s_vname;
    c_vinit =  
      (if typ = GraphPtr && init = C_Noexpr then
        (* Initialize with an empty graph *)
        C_Call("_newGraph", [])
      else
        init);
  }


(* TO CAST In expression: element in list, key in dict, attribute in node *)
and to_c_in_expr (e1: Sast.s_expr) (e2: Sast.s_expr) (t1: Ast.var_type) (t2:
  Ast.var_type) =
  (let e1 = to_c_expr e1 and e2 = to_c_expr e2 in 
  let t1 = to_c_var_type t1 and t2 = to_c_var_type t2 in 
  match t2 with
    Vector(vt) -> C_Call("_contains<" ^ string_c_var_type vt ^ ">", [e2; e1])
  | Map(kt, vt) -> C_Call("_containsKey<" ^ string_c_var_type kt ^ ", "
                          ^ string_c_var_type vt ^ ">", [e2; e1]) 
  | NodePtr -> C_Call("_hasAttribute", [e2; e1]))

(* To CAST Reference: node[attribute], graph[node], dict[key], list[index]*)
and to_c_value (e1: Sast.s_expr) (e2: Sast.s_expr) (t: var_type) =
  let e1 = (to_c_expr e1) and e2 = (to_c_expr e2) in 
    if t = Node then
      C_Value(C_Call("_getAttributes", [e1]), e2)
    else if t = Graph then
      C_Call("_getNode", [e1; e2])
    else
      C_Value(e1, e2)

(* TO CAST expression *)
and to_c_expr (e: Sast.s_expr) : c_expr = match e with  
    S_IntLit(lit) -> C_IntLit(lit)
  | S_FloatLit(lit) -> C_DoubleLit(lit)
  | S_BoolLit(lit) -> C_BoolLit(lit)
  | S_CharLit(lit) -> C_CharLit(lit)
  | S_StrLit(lit) -> C_StrLit(lit)
  | S_Id(id, _) -> C_Id(id)
  | S_Assign(e1, asnop, e2, t) -> to_c_assign e1 asnop e2 t 
  | S_Not(e1, t1) -> to_c_not_expr e1 t1 
  | S_Binop(e1, op, e2, t) ->
      to_c_binop_expr e1 op e2 t
  | S_In(e1, e2, t1, t2) -> to_c_in_expr e1 e2 t1 t2
  | S_Call(fname, params, t) -> 
      C_Call(fname, List.map (fun p -> to_c_expr p) params)
  | S_Value(e1, e2, _, t) -> to_c_value e1 e2 t
  | S_ListLit(el, t) -> to_c_list_lit el t
  | S_DictLit(kvl, kt, vt) -> to_c_dict_lit kvl kt vt
  | S_GraphLit(edgel) -> to_c_graph_lit edgel 
  | S_Null -> C_Null
  | S_Noexpr -> C_Noexpr


(* To CAST Not expression *)
and to_c_not_expr (e1: Sast.s_expr) (t1 : var_type) : c_expr =
  (* Not empty list/dict -> false *)
  if t1 = Void || t1 = List(Void) || t1 = Dict(Void, Void) then
    C_BoolLit(true)
  else
    C_Not(to_c_expr e1)


(* To CAST binary expression *)
and to_c_binop_expr (e1: Sast.s_expr) (op: Ast.op) (e2: Sast.s_expr) (t:
  var_type): c_expr =
  let e1 = to_c_expr e1 and e2 = to_c_expr e2  and t = to_c_var_type t in
  if op = Plus then 
    (match t with
      GraphPtr -> C_Call("_plus", [e1; e2]) 
    | Vector(vt) -> C_Call("_plus<" ^ string_c_var_type vt ^ ">", [e1; e2])
    | Map(kt, vt) -> C_Call("_plus<" ^ string_c_var_type kt ^ ", " ^  
                      string_c_var_type vt ^ ">", [e1; e2])
    | _ -> C_Binop(e1, op, e2)) 
  else (
    if op = Minus && t = GraphPtr then
      C_Call("_minusGraph", [e1; e2])
    else
      C_Binop(e1, op, e2) )

(* TO CAST assign expression *)
and to_c_assign (e1: Sast.s_expr) (asnop: Ast.asnop) (e2: Sast.s_expr)
(t: Ast.var_type) : c_expr =
  let e1 = to_c_expr e1 and e2 = to_c_expr e2 and t = to_c_var_type t in
  match t with
  | Vector(vt) ->
      (match asnop with 
        Asn -> C_Assign(e1, asnop, e2)
      | Add -> C_Call("_plusequal<" ^ string_c_var_type vt ^ ">", [e1; e2])) 
  | Map(kt, vt) ->
      (match asnop with 
        Asn -> C_Assign(e1, asnop, e2)
      | Add -> C_Call("_plusequal<" ^ string_c_var_type vt ^ ", " ^ 
               string_c_var_type kt ^ ">", [e1; e2]) )
  | GraphPtr -> 
      (match asnop with 
        Asn -> C_Assign(e1, asnop, e2)
      | Add -> C_Call("_plusequal",  [e1; e2])
      | Sub -> C_Call("_minusEqualGraph", [e1; e2])) 
  | _  -> C_Assign(e1, asnop, e2)


(* TO CAST list literal: 
  * [|1, 2, 3|] --> ({ vector<int> _tmp = {1, 2, 3}; _tmp; })
  *)
and to_c_list_lit (el: Sast.s_expr list) (t : var_type) =
  let tmp = { 
    c_vtype = Vector(to_c_var_type t);
    c_vname = "_tmp"; 
    c_vinit = C_ListLit(List.map to_c_expr el);
  } in C_Exprstmt([C_Variable(tmp); C_Expr(C_Id("_tmp"))]) 


(* TO CAST list literal: 
 * [|1, 2, 3|] --> ({ vector<int> _tmp = {1, 2, 3}; _tmp; })
 *)
and to_c_dict_lit (kvl: (Sast.s_expr * Sast.s_expr) list) (kt : var_type) (vt:
  var_type) = 
  let tmp = {
    c_vtype = to_c_var_type (Ast.Dict(kt, vt));
    c_vname = "_tmp";
    c_vinit = C_DictLit(List.map (fun kv -> to_c_expr (fst kv), to_c_expr
    (snd kv)) kvl);
  } in C_Exprstmt([C_Variable(tmp); C_Expr(C_Id("_tmp"))])


(* TO CAST graph literal:
 * (| s1--(l1)-->d1; s2--(l2)-->d2 |) ->
 *  _createGraph(2, _createEdge(s1, l1, d1), _createEdge(s2, l2, d2));
 *)
and to_c_graph_lit (edgel: (Sast.s_expr * Sast.s_expr * Sast.s_expr) list) =  
  let el = 
    List.map (fun (src, dest, label) ->
      C_Call("_createEdge",[to_c_expr src; to_c_expr label; to_c_expr dest])) edgel
  in  
    C_Call("_createGraph", C_IntLit(List.length el) :: el) 


(* TO CAST statement *)
and to_c_stmt (s: Sast.s_stmt) = match s with
    S_Block(_, sl) -> C_Block(List.map to_c_stmt sl)
  | S_Variable(var) -> C_Variable(to_c_var_decl var)
  | S_Expr(expr, _) -> C_Expr(to_c_expr expr)
  | S_Return(expr) -> C_Return(to_c_expr expr)
  | S_If(expr, s1, s2) -> C_If(to_c_expr expr, to_c_stmt s1, to_c_stmt s2)
  | S_For(init, test, after, stmt) -> 
      C_For(to_c_expr init, to_c_expr test, to_c_expr after, to_c_stmt stmt)
  | S_Foreach(e1, e2, t, stmt) ->
      if t = Graph then
        (* for (node in graph) -> 
         * for (auto _tmp : graph) { node = _tmp ... } 
         *)
        C_Foreach(C_Id("_tmp"), C_Call("getNodes", [to_c_expr e2]), 
        C_Block([C_Expr(C_Assign(to_c_expr e1, Ast.Asn, C_Id("_tmp")));
        to_c_stmt stmt])) 
      else
      C_Foreach(C_Id("_tmp"), to_c_expr e2, 
      C_Block([C_Expr(C_Assign(to_c_expr e1, Ast.Asn,
      (match t with
         List(_) -> C_Id("_tmp")
        (* for (element in list) -> 
         * for (auto _tmp : list) { element = _tmp ... } 
         *)
       | Dict(_, _) -> C_Id("_tmp.first")
        (* for (key in dict) -> 
         * for (auto _tmp : dict) { key = _tmp ... } 
         *)
      ))) ; to_c_stmt stmt])) 
  | S_While(test, stmt) -> C_While(to_c_expr test, to_c_stmt stmt)
  | S_Continue -> C_Continue
  | S_Break -> C_Break


(* TO CAST function declaration *)
and to_c_func_decl (f: Sast.s_func_decl) =
  { c_fname = f.s_fname;
    c_formals = List.map (fun f -> (to_c_var_type (fst f), snd f)) f.s_formals; 
    c_rtype = to_c_var_type f.s_rtype;
    c_body = List.map to_c_stmt f.s_body; }

(* TO CAST program *)
and to_c_program (program: Sast.s_program) : c_program = 
  { c_global_vars = (List.map to_c_var_decl program.s_var_decls);
    c_func_decls = (List.map to_c_func_decl program.s_func_decls);
    headers = ["<iostream>"; "<string>"; "<vector>"; "<unordered_map>";
               "\"graph.h\""; "\"list.h\""; "\"dict.h\""];
  } 




(********************************************************
 * C++ Code Generation  
 ********************************************************)

let rec string_c_expr (e: c_expr) = match e with
    C_IntLit(lit) -> string_of_int lit
  | C_DoubleLit(lit) -> string_of_float lit
  | C_BoolLit(lit) -> string_of_bool lit
  | C_CharLit(lit) -> "'" ^ Char.escaped lit ^ "'"
  | C_StrLit(lit) -> "\"" ^ lit ^ "\""
  | C_Id(id) -> id
  | C_Assign(e1, asnop, e2) ->
      string_c_expr e1 ^ " " ^ string_of_asnop asnop ^ string_c_expr e2
  | C_Not(e) -> "!(" ^ string_c_expr e ^ ")"
  | C_Binop(e1, op, e2) ->
      string_c_expr e1 ^ " " ^ string_of_binop op ^ " " ^ string_c_expr e2
  | C_Call(fname, params) ->
      fname ^ "(" ^ (String.concat ", " (List.map string_c_expr params)) ^ ")"
  | C_Value(e1, e2) -> string_c_expr e1 ^ "[" ^ string_c_expr e2 ^ "]"
  | C_Null -> "null"
  | C_Noexpr -> "Noexpr"
  | C_Exprstmt(sl) ->
      "({" ^ (String.concat " " (List.map string_c_stmt sl)) ^ "})"
  | C_ListLit(el) ->
      "{" ^ (String.concat ", " (List.map string_c_expr el)) ^ "}" 
  | C_DictLit(kvl) ->
      "{" ^ (String.concat ", " (List.map (fun kv -> "{" ^ string_c_expr
      (fst kv) ^ ", " ^ string_c_expr (snd kv) ^ "}") kvl)) ^ "}" 
  

and string_c_var_decl (v: c_var_decl) =
  let str = string_c_var_type v.c_vtype ^ " " ^ v.c_vname in
  str ^  (if v.c_vinit = C_Noexpr then  ""
          else " = "  ^ string_c_expr v.c_vinit) ^ ";" 

and string_c_if (e: c_expr) (s1: c_stmt) (s2: c_stmt) =
  "if (" ^ string_c_expr e ^ ") " ^ string_c_stmt s1
  ^ (if s2 <> C_Expr(C_Noexpr)  
  then " else " ^ string_c_stmt s2 else "")

and string_c_stmt (s: c_stmt) = match s with 
    C_Block(sl) -> "{\n" ^ (String.concat "\n" (List.map string_c_stmt sl)) ^ "\n}" 
  | C_Variable(var) -> string_c_var_decl var 
  | C_Expr(e) -> string_c_expr e ^ ";"  
  | C_Return(e)  -> "return " ^ string_c_expr e ^ ";"
  | C_If(e, s1, s2) -> string_c_if e s1 s2 
  | C_For (e1, e2, e3, s1) -> 
      "for (" ^ string_c_expr e1 ^ "; " ^ string_c_expr e2 ^ "; " 
      ^ string_c_expr e3 ^ ") " ^ string_c_stmt s1 
  | C_Foreach(e1, e2, s1) -> 
      "for (auto " ^ string_c_expr e1 ^ " : " ^ string_c_expr e2 ^ ") " 
      ^ string_c_stmt s1
  | C_While(e1, s1) -> 
      "while (" ^ string_c_expr e1 ^ ") " ^ string_c_stmt s1 
  | C_Continue -> "continue;"
  | C_Break -> "break;"

 
let string_c_func_decl (f: c_func_decl) = 
  string_c_var_type f.c_rtype ^ " " ^ f.c_fname ^ "("
  ^ (String.concat ", " (List.map (fun f -> (string_c_var_type (fst f))
  ^ " " ^ (snd f)) f.c_formals)) 
  ^ ") {\n" 
  ^ String.concat "\n" (List.map string_c_stmt f.c_body) 
  ^ "\n}"

let string_c_program (program: c_program) =
  (String.concat "\n" (List.map (fun h -> "#include " ^ h) program.headers))
  ^ "\nusing namespace std;\n\n"
  ^ (String.concat "\n" (List.map string_c_var_decl program.c_global_vars))
  ^ "\n\n"
  ^ (String.concat "\n\n" (List.map string_c_func_decl program.c_func_decls))
  ^ "\n\nint main() { __main(); return 0; }"
 
